# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006-2008 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

__maintainer__ = 'Florian Boucault <florian@fluendo.com>'


from elisa.core import plugin_registry

from poblenou_widgets.slideshow import Slideshow
from poblenou_widgets.slideshow_osd import SlideshowOsd
from poblenou_widgets.player_osd import Dock

from pgm.graph.group import Group

import pgm

SlideshowViewClass = plugin_registry.get_component_class('base:slideshow_view')
PlayerViewClass = plugin_registry.get_component_class('poblenou:player_view')

class SlideshowView(SlideshowViewClass, PlayerViewClass):
    default_config = {}
    supported_controllers = ('base:slideshow_controller',)

    SlideshowWidgetClass = Slideshow

    bindings = {'playing':'playing',
                'current_index' : 'index'}

    def __init__(self):
        super(SlideshowView, self).__init__()
        
        # delay before the osd disappears automatically
        self._osd_duration = 3

        # position of the mouse when the dragging started
        self._drag_start_position = None

    def playing__set(self, state):
        if self.controller.focused:
            self.osd_show(self._osd_duration) 

    def index__set(self, index):
        if self.controller.focused:
            self._update_length_and_position()

    def frontend_changed(self, previous_frontend, new_frontend):
        # FIXME: we can't use super() because of context_handle overriding
        #        by both parent views. :(
        
        if new_frontend == None:
            return
        
        viewport = self.frontend.context.viewport_handle
        canvas = viewport.get_canvas()
        viewport.connect("motion-notify-event", self.mouse_moved)

        if self.slideshow_widget == None:
            canvas = self.frontend.context.viewport_handle.get_canvas()
            self.slideshow_widget = self.SlideshowWidgetClass(canvas,
                                                              pgm.DRAWABLE_FAR)
            self.slideshow_widget.size = canvas.size
            self.slideshow_widget.visible = True

            self.slideshow_widget.connect('clicked', self.drag_clicked)
            self.slideshow_widget.connect('drag_begin', self.drag_begin)
            self.slideshow_widget.connect('drag_motion', self.drag_motion)
            self.slideshow_widget.connect('drag_end', self.drag_end)


        if self.group == None:
            canvas = self.frontend.context.viewport_handle.get_canvas()

            # FIXME: why is the OSD not in self.group? Because of resizing
            # issue that will be solved when drawable.regenerate is in place
            # in Pigment.

            # create the group containing all the widgets of the player
            self.group = Group(canvas, pgm.DRAWABLE_FAR)
            self.group.visible = True
            self.context_handle = self.group
         
            theme = self.frontend.theme
            self._normal_osd_status_bg = theme.get_media("dock_background")
            self._mouse_osd_status_bg = theme.get_media("dock_background_mouse")
            self._play_button_image = theme.get_media("play_button_mouse")
            self._pause_button_image = theme.get_media("pause_button_mouse")
         
            self._create_background()
            self._create_playpause_button()
            self._create_osd_status_widget()
            self._create_back_widget()

    def _create_background(self):
        self._background = self.slideshow_widget
        self.group.add(self._background)

    def _create_osd_status_widget(self):
        # FIXME: refactorize. only difference with player view is usage
        #        of SlideshowOsd instead of PlayerOsd
        canvas = self.frontend.context.viewport_handle.get_canvas()
        theme = self.frontend.theme
         
        osd_status_bar_bg = theme.get_media("dock_bar_bg")
        osd_status_bar_fg = theme.get_media("dock_bar_fg")
        
        px, py = 0.7, 0.8
        iw, ih = canvas.width, canvas.height
        osd_status_w = iw * px
        osd_status_h = ih * 0.15
        self._osd_status = SlideshowOsd(canvas,
                                     pgm.DRAWABLE_FAR,
                                     10000,
                                     osd_status_w, osd_status_h,
                                     osd_status_bar_bg,
                                     osd_status_bar_fg,
                                     self._normal_osd_status_bg)
        self._osd_status.opacity = 0
        self._osd_status.position = self._osd_status_position
        self._osd_status.visible = True
        self._osd_status.index = 0
        self._osd_status.length = 0
        self._osd_status.connect("clicked", self._osd_status_clicked)

    def attribute_set(self, origin, key, old_value, new_value):
        super(SlideshowView, self).attribute_set(origin, key, old_value,
                                                 new_value)
        if key == 'focused':
            if (new_value, self.frontend.context.touchscreen) == (True, True):
                 self.osd_show(self._osd_duration)

    def osd_show(self, time_visible=-1):
        self._update_play_button('')
        self._update_length_and_position()
        self._update_osd_layout()
        self._osd_back_button.show(time_visible)
        self._osd_status.show(time_visible)
        self._osd_play_button.show(time_visible)

    def osd_hide(self):
        self._osd_back_button.hide()
        self._osd_status.hide()
        self._osd_play_button.hide() 

    def focused_changed(self, new_value):
        if not new_value:
            self.osd_hide()

    def _update_length_and_position(self):
        self._osd_status.length = len(self.controller.model.playlist)
        self._osd_status.index = self.controller.model.current_index +1

    def _update_play_button(self, state):
        if self.controller.model.playing == True:
            icon = self._pause_button_image
        else:
            icon = self._play_button_image
  
        self._osd_play_button.image_from_path(icon)
        
    def mouse_moved(self, viewport, event):
        if not self.controller.focused:
            return False

        self.osd_show(self._osd_duration)
        return True

    def _osd_play_button_clicked(self, drawable, x, y, z, button, time):
        model = self.controller.model
        model.playing = not model.playing
        self.controller.next_image()
        self._update_play_button('')
        
    def _osd_back_button_clicked(self, drawable, x, y, z, button, time):
        if self.controller is None or not self.controller.focused:
            return False

        self.osd_hide()
        self.controller.model.playling = False
        self.controller.parent.focus()
        return True

    def drag_begin(self, drawable, x, y, z, button, time):
        if self.controller is None or not self.controller.focused:
            return False

        self._drag_start_position = (x, y)
        self.slideshow_widget.move(0.0)
        self.osd_show(self._osd_duration)
        return True
           
    def drag_end(self, drawable, x, y, z, button, time):
        if self.controller is None or not self.controller.focused:
            return False
        
        if self._drag_start_position != None:
            dx = self._drag_start_position[0] - x
            dy = self._drag_start_position[1] - y

            if dx > 0.3:
                if not self.controller.next_image():
                    self.slideshow_widget.release()
            elif dx < -0.3:
                if not self.controller.previous_image():
                    self.slideshow_widget.release()
            else:
                self.slideshow_widget.release()
            
            self._drag_start_position = None
            
            self.osd_show(self._osd_duration)
            
            return True

        return False

    def drag_motion(self, drawable, x, y, z, button, time):
        if self.controller is None or not self.controller.focused:
            return False
        
        if self._drag_start_position != None:
            dx = x - self._drag_start_position[0]
            self.slideshow_widget.move(dx)
            return True

        return False

    def drag_clicked(self, drawable, x, y, z, button, time):
        if self.controller is None or not self.controller.focused:
            return False

        if self.slideshow_widget.visible == False or self.slideshow_widget.opacity == 0.0:
            return False

        if self._drag_start_position != None:
            dx = self._drag_start_position[0] - x
            if dx < -0.05 or dx > 0.05:
                return True
        
        self.osd_show(self._osd_duration)
        return True
