# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006-2008 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

"""
LocalMediaProvider component class
"""


__maintainer__ = 'Florian Boucault <florian@fluendo.com>'
__maintainer2__ = 'Philippe Normand <philippe@fluendo.com>'

from elisa.base_components.media_provider import MediaProvider, UriNotMonitorable
from elisa.base_components.media_provider import NotifyEvent
from elisa.core import common
from elisa.core.media_uri import MediaUri
from elisa.core.media_file import MediaFile
from elisa.core.utils import misc, sorting
from elisa.extern import path

from twisted.internet import reactor

try:
    from elisa.extern.coherence import inotify
except ImportError:
    inotify = None

import os
import shutil
import re

from twisted.internet import defer, task, threads


class LocalMedia(MediaProvider):
    """
    This class implements local filesystem support
    """

    hidden_fnmatch_pattern = '[!.]*'
    
    default_config = {'hidden_file_chars': '!.'}


    def __init__(self):
        MediaProvider.__init__(self)
        self._paths = {}

    def initialize(self):
        ret = super(LocalMedia, self).initialize()
        self._hidden_file_chars = self._get_option('hidden_file_chars')

        return ret

    def _get_option(self, name):
        try:
            return self.config[name]
        except KeyError:
            return self.default_config[name]

    def supported_uri_schemes__get(self):
        return { 'file': 0 }

    def get_media_type(self, uri):
        if os.path.isdir(uri.path):
            return defer.succeed({'file_type': 'directory', 'mime_type': ''})

        metadata_manager = common.application.metadata_manager
        media_db = common.application.media_manager.media_db
        
        if media_db:
            infos = media_db.get_media_information(uri, extended=False)
            if infos:
                file_type = infos.format
                if infos.typ:
                    mime_type = infos.typ
                else:
                    mime_type = None
                media_type = {'file_type': file_type, 'mime_type': mime_type}
                self.debug('returning cached media type %s for uri %s',
                    media_type, uri)
                return defer.succeed({'file_type': file_type,
                        'mime_type': mime_type})

        def get_metadata_done(metadata):
            self.debug("got metadata for %s: %s", uri, metadata)

            del metadata['uri']

            return metadata

        def get_metadata_failure(failure):
            self.debug("error getting media type for uri %s: %s", uri, failure)

            return failure

        req_metadata = {'uri': uri, 'file_type': None, 'mime_type': None}
        get_metadata_dfr = metadata_manager.get_metadata(req_metadata)
        get_metadata_dfr.addCallbacks(get_metadata_done, get_metadata_failure)

        return get_metadata_dfr

    def is_directory(self, uri):
        return defer.succeed(os.path.isdir(uri.path))
    
    def has_children_with_types(self, uri, media_types):
        if not os.path.isdir(uri.path):
            return defer.succeed(False)

        try:
            contents = os.listdir(uri.path)
        except OSError, error:
            self.info("Could not list directory %r: %s", path, error)
            return defer.succeed(False)
        
        if len(contents) == 0:
            return defer.succeed(False)

        media_manager = common.application.media_manager

        # we go over all the children until we find one that matches media_types
        has_children = False

        def get_media_type_done(metadata):
            if not media_types or metadata['file_type'] in media_types:
                global hash_children
                has_children = True

        def get_media_type_failure(failure):
            # consume it
            return None

        def children_iter():
            for child in contents:
                if has_children:
                    break

                if len(child) == 0 or \
                        child[0] in self._hidden_file_chars:
                    yield None
                    continue

                try:
                    child_uri = uri.join(child)
                except UnicodeDecodeError, e:
                    self.warning(e)
                    # skip this child
                    yield None
                    continue

                get_media_type_dfr = media_manager.get_media_type(child_uri)
                get_media_type_dfr.addCallbacks(get_media_type_done,
                        get_media_type_failure)

                yield get_media_type_dfr
        
        def children_iter_done(iterator):
            return has_children

        dfr = task.coiterate(children_iter())
        dfr.addCallback(children_iter_done)

        return dfr

    def get_direct_children(self, uri, list_of_children):
        if not os.path.isdir(uri.path):
            return list_of_children

        contents = os.listdir(uri.path)

        def natural_sort_done(result):
            def children_iter():
                folders = []
                files = []

                for filename in contents:
                    if len(filename) == 0 or \
                            filename[0] in self._hidden_file_chars:
                        yield None
                        continue

                    file_path = os.path.join(uri.path, filename)
                    
                    parts = {'scheme': 'file', 'path': file_path}
                    file_uri = MediaUri(parts)
                    metadata = {}
                    try:
                        metadata['fs_mtime'] = os.stat(file_path).st_mtime
                    except OSError:
                        pass
                    
                    item = (file_uri, metadata)

                    if os.path.isdir(file_path):
                        folders.append(item)
                    else:
                        files.append(item)

                    yield None

                for lst in (folders, files):
                    for item in lst:
                        list_of_children.append(item)
                        yield None

            def children_iter_done(iterator):
                return list_of_children
            
            dfr = task.coiterate(children_iter())
            dfr.addCallback(children_iter_done)

            return dfr

        # sort directory contents alpha-numerically
        dfr = threads.deferToThread(sorting.natural_sort, contents)
        dfr.addCallback(natural_sort_done)

        return dfr

    def _blocking_open(self, uri, mode='r'):
        ret = None
        if not os.path.isdir(uri.path):
            try:
                handle = open(uri.path, mode)
            except Exception, exc:
                self.info("Could not open %s : %s" % (uri, exc))
            else:
                ret = MediaFile(self, handle)
        return ret

    def next_location(self, uri, root=None):
        return threads.deferToThread(self._blocking_next_location, uri,
                                   root=root)

    def _blocking_next_location(self, uri, root=None):
        next_uri = None

        local = uri.path

        if os.path.isdir(local):
            dir_path = local
        else:
            dir_path = os.path.dirname(local)

        root_path = None
        walk_gen = None
        if root:
            root_path = root.path
            walk_gen = self._paths.get(root_path)
        else:
            for walker_root, walker in self._paths.iteritems():
                if local.startswith(walker_root):
                    walk_gen = walker
                    root_path = walker_root
                    break

        if not root_path:
            root_path = dir_path

        end = False
        if not walk_gen:
            p = path.path(root_path)
            walk_gen = p.walk(self.hidden_fnmatch_pattern, errors='ignore')
            self._paths[root_path] = walk_gen
            if root_path != local:
                try:
                    n = walk_gen.next()
                except StopIteration:
                    end = True
                else:
                    while n != local:
                        try:
                            n = walk_gen.next()
                        except StopIteration:
                            del self._paths[root_path]
                            end = True
                            break

        if not end:
            try:
                next_file = walk_gen.next()
            except StopIteration:
                del self._paths[root_path]
                next_file = None

            if next_file:
                parts = {'scheme': 'file', 'path': next_file}
                next_uri = MediaUri(parts)

        self.debug("Next URI of %s: %s", uri, next_uri)
        return next_uri

    def previous_location(self, uri):
        return threads.deferToThread(self._blocking_previous_location, uri)

    def _blocking_previous_location(self, uri):
        if uri == None:
            return None

        if os.path.isdir(uri.path):
            dir_path = uri.path
        else:
            dir_path = os.path.dirname(uri.path)

        p = path.path(dir_path)
        walk_gen = p.walk(self.hidden_fnmatch_pattern, errors='ignore')

        p = dir_path
        if dir_path != uri.path:
            while True:
                try:
                    n = walk_gen.next()
                    if n == uri.path:
                        break
                    else:
                        p = n
                except StopIteration:
                    return None

        if p == uri.path:
            return None
        if p:
            parts = {'scheme': 'file', 'path': p}
            return MediaUri(parts)
        return None

    def monitor_uri(self, uri, callback, *extra_args):

        def inotify_cb(iwp, filename, mask, parameter):
            media_uri = MediaUri(unicode("file://%s/%s" % (iwp.path, filename)))
            event = self._inotify_event_type(mask)
            return callback(media_uri, {}, event,uri, *extra_args)

        if self.uri_is_monitorable(uri):
            try:
                inotify_instance = inotify.INotify()
            except SystemError, error:
                self.warning(error)
                raise UriNotMonitorable(uri)
            else:
                path = uri.path.encode('utf-8')

                self.info("Starting to monitor %r", path)
                try:
                    inotify_instance.watch(path, auto_add = False, \
                                           recursive=False, \
                                           mask = inotify.IN_CREATE \
                                           | inotify.IN_DELETE \
                                           | inotify.IN_ATTRIB \
                                           | inotify.IN_MOVED_TO \
                                           | inotify.IN_MOVED_FROM, \
                                           callbacks=(inotify_cb, None))
                except IOError, error:
                    raise UriNotMonitorable(uri)
        else:
            raise UriNotMonitorable(uri)

    def _inotify_event_type(self, mask):
        if mask & inotify.IN_CREATE or mask & inotify.IN_MOVED_TO:
            return NotifyEvent.ADDED
        elif mask & inotify.IN_DELETE or mask & inotify.IN_UNMOUNT or mask & inotify.IN_MOVED_FROM:
            return NotifyEvent.REMOVED
        elif mask & inotify.IN_ATTRIB:
            return NotifyEvent.MODIFIED
        else:
            # FIXME: exception should be raised
            inotify_instance = inotify.INotify()
            human = inotify_instance.flag_to_human(mask)
            self.info("Unexpected inotify event: %s" % human)
            return NotifyEvent.REMOVED


    def unmonitor_uri(self, uri):
        self.debug("Stopping %r monitoring", uri)
        inotify.INotify().ignore(uri.path)


    def uri_is_monitorable(self, uri):
        monitorable = False
        home = os.path.expanduser('~')
        if inotify != None:
            inotify_instance = inotify.INotify()
            try:
                inotify_instance.watch(home)
            except UriNotMonitorable:
                monitorable = False
            else:
                monitorable = True
                inotify_instance.ignore(home)
        return monitorable

    def uri_is_monitored(self, uri):
        return inotify and inotify.INotify().is_watched(uri.path)

    def copy(self, orig_uri, dest_uri, recursive=False):
        try:
            if recursive:
                shutil.copytree(orig_uri.path, dest_uri.path)
            else:
                shutil.copy(orig_uri.path, dest_uri.path)
        except Exception, e:
            raise e

    def move(self, orig_uri, dest_uri):
        try:
            shutil.move(orig_uri.path, dest_uri.path)
        except shutil.Error, e:
            raise e


    def delete(self, uri, recursive=False):
        try:
            if recursive:
                shutil.rmtree(uri.path)
            else:
                os.remove(uri.path)
        except shutil.Error, e:
            raise e

