# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006-2008 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 2.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

import gobject
gobject.threads_init()
import pygst
pygst.require('0.10')
import gst
import os
import sys
import time
from threading import Lock
import platform

from elisa.base_components.metadata_provider import MetadataProvider
from elisa.core.log import Loggable
from elisa.core.media_uri import MediaUri
from elisa.core.utils import locale_helper
from elisa.core.component import InitializeFailure

from twisted.internet import defer, reactor, error
from twisted.internet.protocol import ProcessProtocol
from twisted.internet.stdio import StandardIO
from twisted.spread import pb, flavors, jelly
from threading import Lock

import pkg_resources

import PIL
from PIL import PngImagePlugin
import Image
import ImageStat
import md5

if platform.system() == 'windows':
    import win32process

SEEK_SCHEDULED = 'scheduled'
SEEK_DONE = 'done'

THUMBNAIL_DIR = os.path.join(os.path.expanduser("~"), ".thumbnails", 'large')
THUMBNAIL_SIZE = 256
BORING_IMAGE_VARIANCE=2000

__maintainer__ = 'Alessandro Decina <alessandro@fluendo.com>'
    
supported_metadata_keys = set(['artist', 'album', 'song', 'track'])
media_type_keys = set(['uri', 'file_type', 'mime_type'])
supported_keys = supported_metadata_keys.union(media_type_keys)

class TimeoutError(Exception):
    def __init__(self):
        Exception.__init__(self, 'timeout')
    
def able_to_handle(supported_schemes, supported_keys, metadata):
    uri = metadata.get('uri')
    if not uri or uri.scheme not in supported_schemes:
        return False

    request_keys = supported_keys.intersection(metadata.keys())
    request_empty_keys = \
            [key for key in request_keys if metadata[key] is None]

    if request_empty_keys:
        return True

    return False 

class GstMetadataPipeline(Loggable):
    reuse_elements = False
    timeout = 2
    thumb_timeout = 2
    
    def __init__(self):
        super(GstMetadataPipeline, self).__init__()
        self._pipeline = None
        self._decodebin = None
        self._ffmpegcolorspace = None
        self._plugged_elements = []
        self._frame_locations = [1.0 / 3.0, 2.0 / 3.0, 0.1, 0.9, 0.5]
        self._probe_lock = Lock()
        self._thumb_probe_id = None

        # other instance variables that need to be reset for each new metadata
        # request are set directly in _reset()

    def clean(self):
        self._clean_pipeline(finalize=True)

        if self._timeout_call is not None:
            self._timeout_call.cancel()
            self._timeout_call = None

        if self._seek_call is not None:
            self._seek_call.cancel()
            self._seek_call = None

    def initialize(self):
        self._reset()

    def _clean_pipeline(self, finalize=False):
        if not self.reuse_elements or finalize:
            # destroy the pipeline
            if self._pipeline is not None:
                self._pipeline.set_state(gst.STATE_NULL)
                self._pipeline = None
                self._decodebin = None
                self._ffmpegcolorspace = None
                self._plugged_elements = []
        else:
            if self._pipeline is not None:
                self._bus.set_flushing(True)
                self._pipeline.set_state(gst.STATE_READY)

            if self._decodebin is not None:
                self._typefind.unlink(self._decodebin)
                self._decodebin.set_state(gst.STATE_NULL)
                self._pipeline.remove(self._decodebin)
                self._decodebin = None
        
            for element in self._plugged_elements:
                self._pipeline.remove(element)
                element.set_state(gst.STATE_NULL)
            self._plugged_elements = []

    def _build_pipeline(self):
        self._pipeline = gst.Pipeline()
        self._bus = self._pipeline.get_bus()
        self._bus.add_signal_watch()
        self._bus.connect('message::application',
                self._bus_message_application_cb)
        self._bus.connect('message::error', self._bus_message_error_cb)
        self._bus.connect('message::eos', self._bus_message_eos_cb)
        self._bus.connect('message::tag', self._bus_message_tag_cb)
        self._bus.connect('message::state-changed',
                self._bus_message_state_changed_cb)
        self._src = gst.element_factory_make('filesrc')
        self._typefind = gst.element_factory_make('typefind')
        self._typefind.connect('have-type', self._typefind_have_type_cb)
        pad = self._typefind.get_pad('src')
        self._pipeline.add(self._src, self._typefind)
        self._src.link(self._typefind)
        
        self._pipeline.set_state(gst.STATE_READY)

    def _reset(self):
        # NOTE: we call gst_element_set_state so we MUST NOT be called from the
        # streaming thread

        self._probe_lock.acquire()
        if self._thumb_probe_id is not None:
            self._rgb_pad.remove_buffer_probe(self._thumb_probe_id)
            self._thumb_probe_id = None
        self._probe_lock.release()

        # destroy the current pipeline if reuse_elements == False, otherwise
        # clean it so that it can be reused
        self._clean_pipeline()
        if self._pipeline is None:
            # we're either being called from initialize() or
            # self.reuse_elements == False
            self._build_pipeline()

        # the metadata dictionary of the current request
        self._req_metadata = None
        # the uri value in the metadata dictionary
        self._req_uri = None
        # the deferred that we callback when we finish loading stuff in
        # self._req_metadata
        self._req_defer = None

        # the caps as given by the typefind::have-type signal
        self._typefind_caps = None
        self._typefind_file_type = None
        self._typefind_mime_type = None

        # the video/audio/image caps that we get from decodebin pads when
        # we plug decodebin
        self._video_caps = None
        self._audio_caps = None
        self._image_caps = None

        # the taglist containing all the tags for the stream
        self._tags = gst.TagList()

        # the duration of the current stream, used to seek when doing a
        # thumbnail 
        self._duration = None
        # needs to be guarded with self._probe_lock
        self._seek_status = None
        self._seek_location_index = 0
        self._seek_call = None

        self._timeout_call = None

        # timestamps used for logging purposes
        self._start_timestamp = 0
        self._end_timestamp = 0

    def _bus_message_error_cb(self, bus, message):
        gerror, debug = message.parse_error() 
        if self._typefind_file_type is not None or \
                self._video_caps is not None or \
                self._audio_caps is not None or \
                self._image_caps is not None:
            # we got an error going to PAUSED but we still can report the info
            # that we got from have_type_cb
            self.debug('error going to paused %s: %s', gerror.message, debug)
            self._done()
        else:
            self._failed(Exception('%s: %s' % (gerror.message, debug)))

    def _bus_message_application_cb(self, bus, message):
        if message.structure.get_name() == 'metadata-done':
            self._done()
            return

        error = Exception(message.structure['error'])
        self._failed(error)

    def _bus_message_eos_cb(self, bus, message):
        self._done()

    def _bus_message_tag_cb(self, bus, message):
        taglist = message.parse_tag()
        self._tags = self._tags.merge(taglist, gst.TAG_MERGE_APPEND)
   
    def _bus_message_state_changed_cb(self, bus, message):
        if message.src == self._pipeline:
            self.debug('state changed %s' % message)

            prev, current, pending = message.parse_state_changed()
            if current == gst.STATE_PAUSED and \
                self._decodebin and self._decodebin.get_pad('sink').is_linked() and \
                (self._typefind_file_type != 'video' or self._have_video_thumbnail()):
                self._done()

    def _typefind_have_type_cb(self, typefind, probability, caps):
        self.debug('have type %s' % caps)

        # self._typefind_caps = caps is broken, bug in the bindings
        # FIXME: fix the bug and change this asap
        self._typefind_caps = caps.copy()
        gst_mime_type = self._typefind_mime_type = self._typefind_caps[0].get_name()
        file_type = self._typefind_file_type = gst_mime_type.split('/')[0]

        # NB: id3 tags most of the time are used with mp3 (even if it isn't
        # uncommon to find them with AIFF or WAV). Given that mp3 is by far the
        # most used audio format at the moment we make the common case fast here
        # by assuming that the file_type is audio. By doing this we also set the
        # mime_type to application/x-id3, but this doesn't matter at the moment
        # since we don't use the mime_type anywhere.
        if gst_mime_type == 'application/x-id3':
            file_type = self._typefind_file_type = 'audio'

        req_keys = set(self._req_metadata.keys())
        if req_keys == media_type_keys and (file_type in ('text', 'audio', 'image') or
                (file_type == 'video' and self._have_video_thumbnail())):
            self.debug('got media_type for %s, NOT going to paused',
                    self._req_uri)
            # we are in the streaming thread so we post a message on the bus
            # here and when we read it from the main thread we call _done()
            structure = gst.Structure('metadata-done')
            self._bus.post(gst.message_new_application(self._pipeline, structure))
            return

        # we need tags and/or a thumbnail
        self.debug('we need to go to PAUSED, plugging decodebin '
                '(file_type: %s, have_thumbnail: %s)', file_type,
                self._have_video_thumbnail())
        self._plug_decodebin()

    def _plug_decodebin(self):
        if self._decodebin is None:
            self._decodebin = gst.element_factory_make('decodebin')
            self._decodebin.connect('new-decoded-pad',
                    self._decodebin_new_decoded_pad_cb)
            self._decodebin.connect('unknown-type',
                    self._decodebin_unknown_type_cb)
            self._pipeline.add(self._decodebin)
        
        self._typefind.link(self._decodebin)
        pad = self._typefind.get_pad('src')
        self._decodebin.set_state(gst.STATE_PAUSED)

    def _get_thumbnail_location(self, uri):
        thumbnail_filename = md5.new(str(uri)).hexdigest() + ".png"

        return os.path.join(THUMBNAIL_DIR, thumbnail_filename)
        
    def _save_thumbnail(self, thumbnail):
        thumbnail_filename = self._get_thumbnail_location(self._req_uri)
        
        directory = os.path.dirname(thumbnail_filename)
        if not os.path.exists(directory):
            try:
                os.makedirs(directory, 0700)
            except OSError, e:
                msg = "Could not make directory %r: %s. Thumbnail not saved." % (directory, e)
                self.warning(msg)
                raise ThumbnailerError(self._req_uri, msg)
                
        info = PngImagePlugin.PngInfo()

        # required metadata
        info.add_text("Thumb::URI", str(self._req_uri))
        thumbnail.save(thumbnail_filename, "png", pnginfo=info)

    def _rgb_pad_probe_cb(self, pad, buffer):
        # Seek forward in the stream a number of times until we find a
        # non-boring frame to thumbnail. Every time we get a boring frame we
        # return False so the buffer is discarded and the pipeline doesn't go to
        # PAUSED

        self._probe_lock.acquire()
        if self._seek_status is None:
            self.debug('first probe, seeking to first location')
            # set _seek_status and schedule a delayed call
            self._seek_next_thumbnail_location()
            self._probe_lock.release()

            return False

        if self._seek_status != SEEK_DONE:
            # we're seeking to a better location, ignore buffers until the seek
            # is done
            self.debug('skipping buffer, seek status %s' % self._seek_status)
            self._probe_lock.release()

            return False
        else:
            self._probe_lock.release()

        caps = pad.props.caps
        if caps is None:
            self.warning('pad %s has no caps, not doing thumbnail' % pad)

            # we need the caps to get width and height of the frame but
            # something is wrong with upstream elements, we don't do the
            # thumbnail and let the sink preroll. This should never happen but
            # seems to happen sometimes with ffmpeg on windows.
            return True

        caps = caps[0]
        width = caps['width']
        height = caps['height']

        # we use PIL to compute the variance so while we are at it, we use PIL
        # to do the thumbnail as well. If we depend on PIL just for this we
        # should probably compute the variance ourselves and drop the
        # dependency.
        try:
            img = Image.frombuffer("RGB", (width, height),
                    buffer, "raw", "RGB", 0, 1)
        except Exception, e:
            self.debug("Invalid frame: %s", e)
            
            return False
        
        stat = ImageStat.Stat(img)
        boring = True
        for i in stat.var:
            if i > BORING_IMAGE_VARIANCE:
                boring = False
                break

        # FIXME: we do a thumbnail if the frame isn't boring or if this is our
        # last seek location. This is suboptimal as we should rather store all the
        # boring thumbnails and return the least boring as a last resort.
        if not boring or \
                self._seek_location_index == len(self._frame_locations):
            if not boring:
                self.debug("non-boring frame found")
            else:
                self.debug('boring frame at the last seek location... '
                        'doing the thumbnail anyway')

            img.thumbnail((THUMBNAIL_SIZE, THUMBNAIL_SIZE), Image.BILINEAR)
            if img.mode != 'RGBA':
                img = img.convert(mode='RGBA')

            self._save_thumbnail(img)

            self._probe_lock.acquire()
            try:
                self._rgb_pad.remove_buffer_probe(self._thumb_probe_id)
                self._thumb_probe_id = None
            finally:
                self._probe_lock.release()

            return True

        else:
            self.debug("boring frame, skipping")
            self._seek_next_thumbnail_location()
            return False

        return False
    
    def _plug_thumbnailbin(self, video_pad):
        if self._ffmpegcolorspace is None:
            self._ffmpegcolorspace = gst.element_factory_make('ffmpegcolorspace')
            self._capsfilter = gst.element_factory_make('capsfilter')
            self._pipeline.add(self._ffmpegcolorspace, self._capsfilter)
            self._capsfilter.props.caps = \
                    gst.Caps('video/x-raw-rgb, bpp=24, depth=24')
            self._ffmpegcolorspace.link(self._capsfilter)
            self._rgb_pad = self._capsfilter.get_pad('src')
            self._ffmpegcolorspace.set_state(gst.STATE_PAUSED)
            self._capsfilter.set_state(gst.STATE_PAUSED)
        video_pad.link(self._ffmpegcolorspace.get_pad('sink'))
        sink = gst.element_factory_make('fakesink')
        self._pipeline.add(sink)
        self._rgb_pad.link(sink.get_pad('sink'))
        sink.set_state(gst.STATE_PAUSED)
        self._plugged_elements.append(sink)
        self._thumb_probe_id = self._rgb_pad.add_buffer_probe(self._rgb_pad_probe_cb)
        
        return self._rgb_pad

    def _have_video_thumbnail(self):
        return os.path.exists(self._get_thumbnail_location(self._req_uri))

    def _find_decoder(self, pad):
        target = pad.get_target()
        element = target.get_parent()
        klass = element.get_factory().get_klass()
        if 'Decoder' in klass:
            return element
        return None

    def _get_type_from_decoder(self, decoder):
        klass = decoder.get_factory().get_klass()
        parts = klass.split('/', 2)
        if len(parts) != 3:
            return None

        return parts[2].lower()

    def _seek_next_thumbnail_location(self):
        self._seek_status = SEEK_SCHEDULED

        self._seek_call = \
            reactor.callLater(0, self._seek_next_thumbnail_location_real)

    def _seek_next_thumbnail_location_real(self):
        self._seek_call = None
        self._probe_lock.acquire()
        self._seek_status = SEEK_DONE
        self._probe_lock.release()

        if self._duration is None:
            # first seek, get the duration
            try:
                self._duration, format = self._pipeline.query_duration(gst.FORMAT_TIME)
            except gst.QueryError, e:
                self.debug('duration query failed: %s', e)
                
                return

            if self._duration == -1:
                self.debug('invalid duration, not seeking')
                return
           
            self.debug('stream duration %s' % self._duration)
        
        if self._seek_location_index == len(self._frame_locations):
            self.debug('no more seek locations')
            return self._failed(Exception('no more seek locations'))

        location = self._frame_locations[self._seek_location_index]
        self.debug('seek to location %d, time %s duration %s' %
                (self._seek_location_index,
                gst.TIME_ARGS(int(location * self._duration)),
                self._duration))
        self._seek_location_index += 1
        
        res = self._pipeline.seek(1.0, gst.FORMAT_TIME,
                gst.SEEK_FLAG_FLUSH | gst.SEEK_FLAG_KEY_UNIT,
                gst.SEEK_TYPE_SET, int(location * self._duration),
                gst.SEEK_TYPE_NONE, 0)
        
        self.debug('seek done res %s' % res)

    def _close_pad(self, pad):
        queue = gst.element_factory_make('queue')
        # set the queue leaky so that if we take some time to do the thumbnail
        # the demuxer doesnt' block on full queues
        queue.props.leaky = 1
        sink = gst.element_factory_make('fakesink')
        self._pipeline.add(queue, sink)
        # add sink before queue so when we iterate over the elements to clean
        # them we clean the sink first and unblock the queue if it's blocked
        # prerolling
        self._plugged_elements.append(sink)
        self._plugged_elements.append(queue)
        pad.link(queue.get_pad('sink'))
        queue.link(sink)
        queue.set_state(gst.STATE_PAUSED)
        sink.set_state(gst.STATE_PAUSED)

    def _get_pad_type(self, pad):
        decoder = self._find_decoder(pad)
        if decoder:
            return self._get_type_from_decoder(decoder)
        
        return pad.get_caps()[0].get_name().split('/', 1)[0]

    def _get_pad_caps(self, pad):
        decoder = self._find_decoder(pad)
        if decoder:
            return decoder.get_pad('sink').get_caps()

        return pad.get_caps()

    def _decodebin_new_decoded_pad_cb(self, decodebin, pad, is_last):
        self.debug('new decoded pad %s, caps %s, is_last %s' % (pad,
                pad.get_caps(), is_last))

        typ = self._get_pad_type(pad)
        caps = self._get_pad_caps(pad)

        if typ == 'audio':
            if self._audio_caps is None:
                self._audio_caps = caps
        elif typ == 'video':
            if self._video_caps is None:
                self._video_caps = caps
                # do a thumbnail of the first video track
                self._plug_thumbnailbin(pad)
        elif typ == 'image':
            if self._image_caps is None:
                self._image_caps = caps

        if not pad.is_linked():
            self._close_pad(pad)
    
    def _decodebin_unknown_type_cb(self, decodebin, pad, caps):
        self.debug('unknown pad %s, caps %s' % (pad, caps))

    def get_metadata(self, requested_metadata):
        assert self._timeout_call is None

        self._req_metadata = requested_metadata
        self._req_uri = requested_metadata['uri']
        self._req_defer = defer.Deferred()

        self.debug('getting metadata %s' % self._req_metadata)

        self._start_timestamp = time.time()
        system_encoding = locale_helper.system_encoding()
        self._src.props.location = \
                self._req_metadata['uri'].path.encode(system_encoding)

        self._timeout_call = reactor.callLater(self.timeout, self._timeout)

        # reset the bus in case this is not the first request  
        self._bus.set_flushing(False)
        self._pipeline.set_state(gst.STATE_PAUSED)

        return self._req_defer

    def _get_media_type_from_caps(self, caps):
        res = {}
        mime_type = caps[0].get_name()
        file_type = mime_type.split('/', 1)[0]

        return {'file_type': file_type, 'mime_type': mime_type}

    def _done(self):
        if not self._timeout_call.called:
            self._timeout_call.cancel()

        # we can't check self._seek_call.called here because we don't know if we
        # scheduled a seek call at all
        if self._seek_call is not None:
            self._seek_call.cancel()
            self._seek_call = None

        self._end_timestamp = time.time()

        metadata = self._req_metadata
        metadata_defer = self._req_defer

        available_metadata = {}
        for caps in (self._video_caps, self._audio_caps,
                self._image_caps):
            if caps is not None:
                available_metadata.update(self._get_media_type_from_caps(caps))
                break

        # fallback to typefind caps
        if available_metadata.get('file_type') is None:
            available_metadata['file_type'] = self._typefind_file_type
            available_metadata['mime_type'] = self._typefind_mime_type

        tags = self._tags

        try:
            del tags['extended-comment']
        except KeyError:
            pass

        tag_keys = tags.keys()
        for gst_key, elisa_key in (('track-number', 'track'),
                    ('title', 'song')):
            try:
                available_metadata[elisa_key] = tags[gst_key]
            except KeyError:
                pass
            
        for key in tag_keys:
            value = tags[key]
            # FIXME: this was an old assumption, let's keep it until we update
            # all the old code
            if isinstance(value, list):
                try:
                    value = value[0]
                except IndexError:
                    continue

            available_metadata[key] = value
        
        for key, value in available_metadata.iteritems():
            try:
                if metadata[key] is None:
                    metadata[key] = value
            except KeyError:
                pass

        self.debug('finished getting metadata %s, elapsed time %s' % 
                (metadata, self._end_timestamp - self._start_timestamp))
        
        self._reset()
        metadata_defer.callback(metadata)

    def _timeout(self, thumb_timeout=False):
        self.debug('timeout thumb %s video caps %s',
                thumb_timeout, self._video_caps)

        if not thumb_timeout and (self._typefind_file_type == 'video' or
                self._video_caps is not None):
            # give some more time to the pipline if we are trying to make a
            # thumbnail
            self._timeout_call = \
                reactor.callLater(self.thumb_timeout, self._timeout, True)
        else:
            if self._typefind_file_type is not None or \
                    self._video_caps is not None or \
                    self._audio_caps is not None or \
                    self._image_caps is not None:
                # timeout while going to paused. This can happen on really slow
                # machines while doing the thumbnail. Even if we didn't do the
                # thumbnail, we have some clue about the media type here.
                self._done()
            else:
                self._failed(TimeoutError())

    def _failed(self, error):
        # cancel delayed calls
        if not self._timeout_call.called:
            self._timeout_call.cancel()

        if self._seek_call is not None:
            self._seek_call.cancel()
            self._seek_call = None
 
        self._end_timestamp = time.time()

        metadata = self._req_metadata
        metadata_defer = self._req_defer
        self.debug('error getting metadata %s, error: %s, '
                'elapsed time: %s, timeout %s' % (metadata, error,
                self._end_timestamp - self._start_timestamp,
                self._timeout_call.called))

        self._reset()
        metadata_defer.errback(error)

class GstMetadata(MetadataProvider):
    _supported_schemes = ['file']
    
    def __init__(self, pipeline=None):
        super(GstMetadata, self).__init__()
        if pipeline is None:
            pipeline = GstMetadataPipeline()
        self._pipeline = pipeline
        self._uri_cache = {}
        self._requests = []
        self._running = False
        self._process_next_call = None
        self._next_scheduled = False
        self._current_metadata = None
        self._current_defer = None
    
    def clean(self):
        if self._process_next_call is not None:
            self._process_next_call.cancel()
            self._process_next_call = None
        self._pipeline.clean()

    def initialize(self):
        try:
            self._pipeline.initialize()
        except Exception, exc:
            msg = "Could not initialize the Pipeline: %s" % exc
            raise InitializeFailure(msg)

    def _reset(self):
        # timestamps used for logging purposes
        self._start_timestamp = 0
        self._end_timestamp = 0
        self._next_scheduled = False

    def able_to_handle(self, metadata):
        return able_to_handle(self._supported_schemes,
                supported_keys, metadata)

    def get_metadata(self, metadata, low_priority=False):
        dfr = defer.Deferred()
        if low_priority:
            self._requests.append((metadata, dfr))
        else:
            self._requests.insert(0, (metadata, dfr))

        if not self._running:
            self._running = True
            self._process_next()

        return dfr

    def _process_next(self):
        assert self._process_next_call is None
        assert not self._next_scheduled
        self._next_scheduled = True

        self._process_next_call = \
                reactor.callLater(0.02, self._process_next_real)

    def _process_next_real(self):
        self._process_next_call = None
        assert self._current_metadata is None
        assert self._current_defer is None

        self._next_scheduled = False

        try:
            self._current_metadata, self._current_defer = self._requests.pop(0)
        except IndexError:
            self.debug('metadata queue empty')
            self._running = False
            return

        self.debug('getting metadata %s, queue length %d' % 
                (self._current_metadata, len(self._requests)))

        self._start_timestamp = time.time()

        metadata_defer = self._pipeline.get_metadata(self._current_metadata)
        metadata_defer.addCallbacks(self._done, self._failed)

    def _done(self, metadata):
        self._end_timestamp = time.time()

        metadata, self._current_metadata = self._current_metadata, None
        metadata_defer, self._current_defer = self._current_defer, None
        
        self._process_next()
        # NOTE: this can't be called from the streaming thread
        self._reset()

        metadata_defer.callback(metadata)

    def _failed(self, failure):
        self._end_timestamp = time.time()
        
        metadata, self._current_metadata = self._current_metadata, None
        metadata_defer, self._current_defer = self._current_defer, None

        self._process_next()
        self._reset()

        metadata_defer.errback(failure)

class MetadataClientProcessProtocol(ProcessProtocol):
    def __init__(self):
        self.broker = pb.Broker(True, jelly.DummySecurityOptions())
        
    def makeConnection(self, transport):
        self.broker.makeConnection(transport)
        ProcessProtocol.makeConnection(self, transport)

    def connectionMade(self):
        ProcessProtocol.connectionMade(self)
        self.broker.connectionMade()

    def outReceived(self, data):
        self.broker.dataReceived(data)
    
    def errReceived(self, data):
        sys.stderr.write(data)

    def processEnded(self, reason):
        ProcessProtocol.processEnded(self, reason)
        self.broker.connectionLost(reason)
        self.factory.processEnded(reason)

class MetadataClientProcessLauncher(Loggable, pb.PBClientFactory):
    max_retries = 3
    server_script = \
            pkg_resources.resource_filename('elisa.plugins.good.gstreamer_plugin',
                    'gst_metadata_runner.py')

    log_category = 'gst_metadata_client_process_launcher'

    def __init__(self, *args, **kw):
        Loggable.__init__(self, *args, **kw)
        pb.PBClientFactory.__init__(self, *args, **kw)
        Loggable.__init__(self)

        self.protocol = MetadataClientProcessProtocol
        self.path = os.path.split(sys.modules['elisa'].__path__[0])[0]
        self.env = dict(os.environ)
        self.env['PYTHONPATH'] = \
                os.pathsep.join([self.env.get('PYTHONPATH', ''), self.path])
        self.args = [sys.executable, '-u', self.server_script]
        self.process = None
        self.retries = 0
        self.start_defer = None
        self.stop_defer = None

    def buildProtocol(self, addr):
        protocol = pb.PBClientFactory.buildProtocol(self, addr)
        protocol.broker.factory = self

        return protocol

    def startProcess(self):
        def get_root_object_done(component):
            self.debug('initializing remote component')
            dfr = component.callRemote('initialize')
            dfr.addCallbacks(initialize_done, initialize_failure,
                callbackArgs=(component,), errbackArgs=(component,))

            return dfr

        def initialize_done(result, component):
            self.info('metadata server started')
            start_defer = self.start_defer
            self.start_defer = None
            res = start_defer.callback(component)

            return res

        def initialize_failure(failure, component):
            self.warning('failed to initialize remote component: %s', failure)
            start_defer = self.start_defer
            self.start_defer = None
            res = start_defer.errback(failure)

            return res

        self.info('starting metadata server')
        
        assert self.start_defer is None
        # start_defer will be called back after initialize() has been called on
        # the remote object
        self.start_defer = defer.Deferred()

        protocol = self.buildProtocol(None)
        self.process = reactor.spawnProcess(protocol,
                sys.executable, self.args, env=self.env)
        
        if platform.system() == 'Windows':
            import win32process
            win32process.SetPriorityClass(self.process.hProcess,
                    win32process.IDLE_PRIORITY_CLASS)

        self.retries = 0
        self.debug('metadata process started')

        # get the remote object and call initialize() on it
        dfr = self.getRootObject()
        dfr.addCallback(get_root_object_done)

        return self.start_defer

    def stopProcess(self):
        assert self.stop_defer is None
        
        self.info('stopping metadata server')

        if self.process is None:
            return dfr.succeed(None)

        def get_component_done(component):
            self.debug('cleaning remote component')
            return component.callRemote('clean')

        def clean_done(component):
            self.debug('clean done')
            self.process.loseConnection()

            return component

        # stop_defer will be fired in processEnded after we terminate the child
        # process
        self.stop_defer = defer.Deferred()

        # get the component and call clean() on it
        dfr = self.get_component()
        # FIXME: handle errback
        dfr.addCallback(get_component_done)
        dfr.addCallback(clean_done)

        return self.stop_defer

    def get_component(self):
        if self.process is None:
            return defer.fail(Exception('process not started'))

        if self.start_defer is not None:
            # the remote object is being started
            return self.start_defer
        else:
            return self.getRootObject()

    # FIXME: this should be called by BaseConnector 
    #def clientConnectionLost(self, connector, reason, reconnecting):

    # MetadataClientProcessProtocol callback
    def processEnded(self, reason):
        self.process = None

        self.log('process terminated %s', reason.getErrorMessage())

#        if reason.type == error.ProcessDone:
        # FIXME: for some reason sometime we get ProcessDone and sometime we get
        # ProcessTerminated when self.process.loseConnection() is called in
        # stopProcess(). The problem is that ProcessTerminated is also used for
        # segfault so we can't check for that here.
        if self.stop_defer is not None:
            self.info('metadata server stopped')
            stop_defer = self.stop_defer
            self.stop_defer = None
            self._failAll(reason)
            if stop_defer:
                stop_defer.callback(None)
            return

        if self.retries == self.max_retries:
            self.info('%d tries done, giving up' % self.retries)
            # FIXME: this should be called from clientConnectionLost but i don't
            # think that that's called at all with the process api...
            self._failAll(reason)
            
            return

        self.retries += 1
        self.startProcess()

class GstMetadataClient(MetadataProvider):
    # FIXME see able_to_handle
    _supported_schemes = ['file']
    _supported_keys = set(['file_type', 'mime_type',
            'artist', 'album', 'song', 'track'])
    
    def __init__(self):
        super(GstMetadataClient, self).__init__()
        self.launcher = MetadataClientProcessLauncher()

    def initialize(self):
        def start_process_done(component):
            return self

        dfr = self.launcher.startProcess()
        dfr.addCallback(start_process_done)

        return dfr

    def clean(self):
        def stop_process_done(result):
            return super(GstMetadataClient, self).clean()

        dfr = self.launcher.stopProcess()
        dfr.addCallback(stop_process_done)
        
        return dfr

    def able_to_handle(self, metadata):
        return able_to_handle(self._supported_schemes,
                supported_keys, metadata)

    def get_metadata(self, metadata, low_priority=False):
        def get_metadata_done(remote_metadata):
            # we get a normal dictionary as the result of a remote get_metadata
            # call but we need to return the metadata dictionary that was passed
            # as argument
            for key, value in remote_metadata.iteritems():
                if value is None:
                    continue
                
                metadata[key] = value

            return metadata

        def got_root(root):
            dfr = root.callRemote('get_metadata', metadata, low_priority)
            dfr.addCallback(get_metadata_done)
            
            return dfr

        root_dfr = self.launcher.get_component()
        root_dfr.addCallback(got_root)

        return root_dfr
