# Elisa - Home multimedia server
# Copyright (C) 2007-2008 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.

__maintainer__ = 'Alessandro Decina <alessandro@fluendo.com>'

import gobject
gobject.threads_init()
import gst

from elisa.core.tests.elisa_test_case import ElisaTestCase
from elisa.core.tests.component_test_case import ComponentTestCase
from elisa.core.media_uri import MediaUri

from twisted.internet import defer, reactor
from twisted.python import failure

from elisa.plugins.good.gstreamer_plugin.gst_metadata import \
        GstMetadataPipeline, MetadataClientProcessLauncher, TimeoutError

class GPropsProxy(object):
    def __init__(self, gprops):
        self.__dict__['gprops'] = gprops

    def __setattr__(self, name, value):
        if name == 'location':
            return

        setattr(self.gprops, name, value)

class TestMetadataAudioPipeline(GstMetadataPipeline):
    def _build_pipeline(self):
        GstMetadataPipeline._build_pipeline(self)
        src = self._src
        src.unlink(self._typefind)
        self._pipeline.remove(src)
        src.set_state(gst.STATE_NULL)
        src = self._src = gst.element_factory_make('audiotestsrc')
        identity = self._identity = gst.element_factory_make('identity')
        src.props = GPropsProxy(src.props)
        self._pipeline.add(src, identity)
        gst.element_link_many(src, identity, self._typefind)
        src.set_state(gst.STATE_READY)
        identity.set_state(gst.STATE_READY)

class TestMetadataVideoPipeline(GstMetadataPipeline):
    def _build_pipeline(self):
        GstMetadataPipeline._build_pipeline(self)
        src = self._src
        src.unlink(self._typefind)
        self._pipeline.remove(src)
        src.set_state(gst.STATE_NULL)
        src = self._src = gst.element_factory_make('videotestsrc')
        identity = self._identity = gst.element_factory_make('identity')
        src.props = GPropsProxy(src.props)
        self._pipeline.add(src, identity)
        gst.element_link_many(src, identity, self._typefind)
        src.set_state(gst.STATE_READY)
        identity.set_state(gst.STATE_READY)
        self._have_thumbnail = False

    def _have_video_thumbnail(self):
        # toggle _have_thumbnail so that we go to PAUSED and then stop on the
        # first frame
        self._have_thumbnail = not self._have_thumbnail
        return not self._have_thumbnail

class TestGstMetadataPipelineCommon(object):
    """Whitebox tests for GstMetadataPipeline"""
    def setUpClass(self):
        self.audio_pipeline = self.build_audio_pipeline()
        self.video_pipeline = self.build_video_pipeline()
        self.audio_pipeline.initialize()
        self.video_pipeline.initialize()

    def tearDownClass(self):
        self.audio_pipeline.clean()
        self.video_pipeline.clean()

    def test_get_audio_media_type(self):
        def get_metadata_done(result):
            self.assertEqual(result['file_type'], 'audio')
        
        media_type_dict = {'uri': MediaUri('file://not_used_for_the_test'),
                'file_type': None, 'mime_type': None}
        dfr = self.audio_pipeline.get_metadata(media_type_dict)
        dfr.addCallback(get_metadata_done)

        return dfr

    def test_get_video_media_type(self):
        media_type_dict = {'uri': MediaUri('file://not_used_for_the_test'),
                'file_type': None, 'mime_type': None}
        dfr = self.video_pipeline.get_metadata(media_type_dict)

        def get_metadata_done(result):
            self.assertEqual(result['file_type'], 'video')

        dfr.addCallback(get_metadata_done)

        return dfr

    def _eat_buffers_probe_cb(self, buffer, data):
        return False

    def test_metadata_timeout(self):
        pad = self.audio_pipeline._src.get_pad('src')
        probe_id = pad.add_buffer_probe(self._eat_buffers_probe_cb)
 
        def get_metadata_result(result):
            pad.remove_buffer_probe(probe_id)
            self.failUnless(isinstance(result, failure.Failure))
            result.trap(TimeoutError)

        media_type_dict = {'uri': MediaUri('file://not_used_for_the_test'),
                'file_type': None, 'media_type': None}
        dfr = self.audio_pipeline.get_metadata(media_type_dict)
        dfr.addBoth(get_metadata_result)
        
        return dfr

    def test_error(self):
        self.audio_pipeline._identity.props.error_after = 1
 
        def get_metadata_result(result):
            self.audio_pipeline._identity.props.error_after = -1
            self.failUnless(isinstance(result, failure.Failure)) 
        
        media_type_dict = {'uri': MediaUri('file://not_used_for_the_test'),
                'file_type': None, 'media_type': None}
        dfr = self.audio_pipeline.get_metadata(media_type_dict)
        dfr.addBoth(get_metadata_result)
        
        return dfr

class TestGstMetadataPipelineNoReuse(ElisaTestCase,
        TestGstMetadataPipelineCommon):

    skip = "Messes up with GIL"
    
    def build_audio_pipeline(self):
        pipeline = TestMetadataAudioPipeline()
        pipeline.reuse_elements = False
        
        return pipeline
    
    def build_video_pipeline(self):
        pipeline = TestMetadataVideoPipeline()
        pipeline.reuse_elements = False
        
        return pipeline

class TestGstMetadataPipelineReuse(ElisaTestCase,
        TestGstMetadataPipelineCommon):

    skip = "Messes up with GIL"
    
    def build_audio_pipeline(self):
        pipeline = TestMetadataAudioPipeline()
        pipeline.reuse_elements = True
        
        return pipeline
    
    def build_video_pipeline(self):
        pipeline = TestMetadataVideoPipeline()
        pipeline.reuse_elements = True
        
        return pipeline

class TestGstMetadataBase(object):
    def test_able_to_handle(self):
        # no uri
        metadata = {'foo': None}
        self.failIf(self.component.able_to_handle(metadata))

        # no uri and supported field
        metadata = {'song': None}
        self.failIf(self.component.able_to_handle(metadata))

        # unsupported uri scheme
        metadata = {'uri': MediaUri('gopher://')}
        self.failIf(self.component.able_to_handle(metadata))

        # unsupported uri scheme with supported empty field
        metadata = {'uri': MediaUri('gopher://'), 'song': None}
        self.failIf(self.component.able_to_handle(metadata))

        # supported uri but no empty field
        metadata = {'uri': MediaUri('file://'), 'artist': 'fabrizio de andre'}
        self.failIf(self.component.able_to_handle(metadata))

        # supported uri but no supported field empty
        metadata = {'uri': MediaUri('file://'), 'artist': 'fabrizio de andre',
                'foo': None}
        self.failIf(self.component.able_to_handle(metadata))
        
        # supported uri and supported field empty
        metadata = {'uri': MediaUri('file://'), 'song': None}
        self.failUnless(self.component.able_to_handle(metadata))

class TestGstMetadata(ComponentTestCase, TestGstMetadataBase):
    component_class = \
            'elisa.plugins.good.gstreamer_plugin.gst_metadata:GstMetadata' 

class TestGstMetadataClient(ComponentTestCase, TestGstMetadataBase):
    component_class = \
            'elisa.plugins.good.gstreamer_plugin.gst_metadata:GstMetadataClient'

class TestMetadataClientProcessLauncher(ElisaTestCase):
    def setUp(self):
        self.launcher = MetadataClientProcessLauncher()

    def test_start_stop(self):
        def get_component_done_after_stop(result):
            self.failUnless(isinstance(result, failure.Failure))

        dfr = self.launcher.startProcess()
        dfr.addCallback(lambda result: self.launcher.stopProcess())
        dfr.addCallback(lambda result: self.launcher.get_component())
        dfr.addBoth(get_component_done_after_stop)

        return dfr
        

    def test_restart(self):
        """Start the metadata_runner process, kill it and check that a
        get_component() call still succeeds
        """
        def second_get_component_done(result, test_dfr):
            dfr = self.launcher.stopProcess()
            dfr.chainDeferred(test_dfr)

        def continue_after_kill(test_dfr):
            dfr = self.launcher.get_component()
            dfr.addCallback(second_get_component_done, test_dfr)
            
        def get_component_done(component):
            self.launcher.process.signalProcess('KILL')
            # continue later so the signal is dispatched
            dfr = defer.Deferred()
            reactor.callLater(0.1, continue_after_kill, dfr)

            return dfr

        dfr = self.launcher.startProcess()
        dfr.addCallback(get_component_done)

        return dfr

    def test_give_up(self):
        """Start the process, set max_retries to 0 and kill the process. Check
        that the launcher gives up after max_retries
        """
        def get_component_result(result, test_dfr):
            self.failUnless(isinstance(result, failure.Failure))
            test_dfr.callback(None)

        def continue_after_kill(test_dfr):
            dfr = self.launcher.get_component()
            dfr.addBoth(get_component_result, test_dfr)
        
            return dfr

        def get_component_done(component):
            self.launcher.max_retries = 0
            self.launcher.process.signalProcess('KILL')
            dfr = defer.Deferred()
            reactor.callLater(0.1, continue_after_kill, dfr)
            
            return dfr 

        dfr = self.launcher.startProcess()
        dfr.addCallback(get_component_done)

        return dfr

